/**
 * GenerationMethods.cpp - Bitcoin address generator
 *
 * Copyright (C)2015 The bitgen developers <bitgen@openmailbox.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
/**
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE. 
*/
 


#include "EntryInfoSet.h"
#include "currencySpecific.h"
#include "bitgenVersion.h"
#include "GenerationInfo.h"
#include "GenerationMethods.h"
#include "OtpFolderHandler.h"
#include "FileUtil.h"
#include "postscriptGeneration.h"
#include "InformationList.h"
#include "bitgenConfig.h"
#include "Timestamp.h"

#include <string>
#include <sstream>

//Pad info generation
std::string InformationList::createOtpListTxtFile(const EntroInfoSet& infoSet, const double sum, const double priceSum, 
	const int padnum, const int totalPads,  const std::string& thetime, const bool listAll,
	int& numEntries)
{	
	std::stringstream ss;	
	
	numEntries = 0;
	
	for(const std::pair<int, EntryInfo>& val : infoSet.getRead())
	{		
		numEntries++;
		ss << minSize(intToString(val.first), 4) << " " << val.second.date << " name: " << minSize(val.second.accountName.substr(0, 19), 20) << 
			((totalPads == 0) ? " addr: " : " pad: ") << val.second.pad << " amount: " << minSize(val.second.amount.toOptionalString(), 10) << " price: " << val.second.price.toOptionalString() << std::endl;
	}

	if(sum != 0.0)
	{
		ss << "Total sum: " << sum << std::endl;
	}

	if(priceSum != 0.0)
	{
		ss << "Price sum: " << priceSum << std::endl;
	}


	std::stringstream ssStart;	

	if(totalPads == 0)
	{
		ssStart << "Public info"  << std::endl;		
	}
	
	ssStart << "Date       : " << thetime << std::endl;
	ssStart << "Currency   : " << getCurrencyName() << std::endl;
	
	ssStart << "Generated  : " << applicationName << " " << versionNumber << std::endl;
	if(totalPads != 0)	
	{
		ssStart << "PadNum     : " << padnum << "/" << totalPads << std::endl;
	}	
	ssStart << "NodeName   : " << getNodeName() << std::endl;
	ssStart << "Num entries: " << numEntries /*info.size()*/ << std::endl;
	
	ssStart << std::endl;

	std::stringstream ssTotal;	
	ssTotal << ssStart.str() << ss.str() << std::endl;


	return ssTotal.str();
}




std::string InformationList::createOtpListPsFile(const EntroInfoSet& infoSet, const std::string& folder, const int padnum, const int totalPads,  const std::string& thetime, const std::string& name,
	const bool listAll)
{	
	const std::string psStr = PostscriptGenerator::generatePostscriptOtpList(infoSet, padnum, totalPads, thetime, name, listAll);
	return psStr;
}




std::string InformationList::createPubOtpListPsFile(const EntroInfoSet& infoSet, const double sum, const double priceSum, const std::string& folder,  const std::string& thetime, const std::string& name, const bool listAll)
{	
	const std::string psStr = PostscriptGenerator::generatePostscriptPubOtpList(infoSet, sum, priceSum, thetime, name, listAll);
	return psStr;	
}
















void InformationList::listPublicFolderContents(const std::string& pubFolder, const std::string& thetime, const std::string& name, const bool listAll)
{
	const std::string totFolder = OtpFolderHandler::getPublicTotFolder(pubFolder);	
	const EntroInfoSet infoSet = OtpFolderHandler::getPublicOtpInfoList(totFolder);


	EntroInfoSet toBePrintedSet;


	for(const std::pair<int, EntryInfo>& val : infoSet.getRead())
	{
		if(!listAll && val.second.isEmptyAmount)
		{
			continue;	//Do not print
		}

		toBePrintedSet.insert(val);
		
	}

	const double sum = toBePrintedSet.sum();
	const double priceSum = toBePrintedSet.priceSum();




	int numListedItems(0);		
	const std::string fileContents = createOtpListTxtFile(toBePrintedSet, sum, priceSum, 0, 0, thetime, listAll, numListedItems);

	const int nextNum = BitgenConfig::getConfiguredNum(false);

	const std::string filenameBase(totFolder + "allpub_" + getCurrencyShortName() + "_" + getCurrentDateString() + "_items_" + intToString(nextNum) + "_pub");

	const std::string filenameTxt = filenameBase + ".txt";
	std::cout << "Writing file: " << filenameTxt << " with " << numListedItems << " entries" << std::endl;	
	fileWrite(filenameTxt, fileContents);	


	const std::string psFileContents = createPubOtpListPsFile(toBePrintedSet, sum, priceSum, totFolder, thetime, name, listAll);

	const std::string filenamePs = filenameBase + ".ps";
	std::cout << "Writing file: " << filenamePs << " with " << numListedItems << " entries" << std::endl;	
	fileWrite(filenamePs, psFileContents);		
}






void InformationList::listFolderContents(const std::string& folder, const int padnum, const int totalPads, const std::string& thetime, const std::string& name, const bool listAll)
{	
	if(!directoryExist(folder))
	{
		std::cout << "No information exist for " << getCurrencyShortName() << " pad: " << padnum << std::endl;
		return;
	}

	const EntroInfoSet infoSet = OtpFolderHandler::getPrivOtpInfoList(folder, padnum);

	EntroInfoSet toBePrintedSet;

	for(const std::pair<int, EntryInfo>& val : infoSet.getRead()) //info)
	{
		if(!listAll && val.second.isEmptyAmount)
		{
			continue;	//Do not print
		}

		toBePrintedSet.insert(val);
	}


	const double sum = toBePrintedSet.sum();
	const double priceSum = toBePrintedSet.priceSum();

	int numListedItems(0);		
	const std::string fileContents = createOtpListTxtFile(toBePrintedSet, sum, priceSum, padnum, totalPads, thetime, listAll, numListedItems);
	const std::string psFileContents = createOtpListPsFile(infoSet, folder, padnum, totalPads, thetime, name, listAll);

	const int nextNum = BitgenConfig::getConfiguredNum(false);

	const std::string filenameBase(folder + "allpad_" + getCurrencyShortName() + "_" + getCurrentDateString() + "_items_" + intToString(nextNum) + "_pad_" + intToString(padnum));

	const std::string filenameTxt = filenameBase + ".txt";
	std::cout << "Writing file: " << filenameTxt << " with " << numListedItems << " entries" << std::endl;	
	fileWrite(filenameTxt, fileContents);

	const std::string filenamePs = filenameBase + ".ps";
	std::cout << "Writing file: " << filenamePs << " with " << numListedItems << " entries" << std::endl;	
	fileWrite(filenamePs, psFileContents);	
}





void InformationList::generateOtpListForPad(const std::list<std::string>& folders, const int numPad, const std::string& thetime, const std::string& name, const bool listAll)
{			
	const int totalPads = folders.size();

	if(numPad > totalPads)
	{
		std::string errStr("Pad num does not exist: ");		
		errStr += intToString(numPad);
		throw errStr;
	}
	
	int currentNum(1);
	for(const std::string& folder : folders)
	{
		if(currentNum == numPad)
		{
			const std::string totFolder = OtpFolderHandler::getPrivPadTotFolder(folder, numPad);

			listFolderContents(totFolder, numPad, totalPads, thetime, name, listAll);
			return;
		}
		currentNum++;
	}
}



