/**
 * bitcoin.cpp - Bitcoin address generator
 *
 * Copyright (C)2015 The bitgen developers <bitgen@openmailbox.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
/**
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE. 
*/
 

#include "bitgenVersion.h"
#include "currencySpecific.h"
#include "util.h"
#include "GenerationMethods.h"
#include "EntityInformationCommandline.h"
#include "license.h"
#include "commandLine.h"
#include "bitgencli.h"
#include "pubXmrAddress.h"
#include "xmrMnemonic.h"
#include "randomUtils.h"
#include "test/unittestMonero.h"
#include "random.h"
#include "bitgenConfig.h"
#include "xmrPrivKey.h"
#include "StaticXor.h"
#include "SaveFolderHandler.h"
#include "language.h"
#include "printPS.h"
#include "FileUtil.h"
#include "keccakHash.h"

#include <iostream>
#include <time.h>
#include <iomanip>



void GenerationMethods::generateXmrFromPrivSpendKey(const BigInt<256>& givenPrivKeyReversed, GenerationParameters& params, const std::string& thetime, 
	const std::string& genMethod, std::list<ExtraInfo>& privateGenerationExtra)
{	
	const BigInt<256>& givenPrivKey(BigInt<256>::fromBytes(reverseStr(givenPrivKeyReversed.getBinaryFull())));
	
	const BigInt<256> privKeyMax = XmrEllipticCurve::getEdwards25519Max(givenPrivKey);

	if(privKeyMax != givenPrivKey)
	{
		std::cout << "Incorrect priv key: " << givenPrivKeyReversed << ", larger than allowed" << std::endl;
		throw std::string("Incorrect priv key");
	}
	
	const XmrPrivSpendKey privSpendKey(givenPrivKey);

	generateXmrFromThePrivSpendkey(privSpendKey, params, thetime, genMethod, privateGenerationExtra);	
}




void GenerationMethods::generateXmrFromRawPrivkey(const BigInt<256>& privKeyRaw, GenerationParameters& params, const std::string& thetime, 
	const std::string& genMethod, std::list<ExtraInfo>& privateGenerationExtra)
{
	const BigInt<256> privKeyFinal = XmrEllipticCurve::getEdwards25519Max(privKeyRaw);

	const XmrPrivSpendKey privSpendKey(privKeyFinal);

	generateXmrFromThePrivSpendkey(privSpendKey, params, thetime, genMethod, privateGenerationExtra);
}



void GenerationMethods::generateXmrFromThePrivSpendkey(const XmrPrivSpendKey& privSpendKey, GenerationParameters& params, const std::string& thetime, 
	const std::string& genMethod, std::list<ExtraInfo>& privateGenerationExtra)
{
	const XmrPrivViewKey privViewKey = privSpendKey.privateSpendToView();
	
	const XmrMnemonic mnemonicObj(privSpendKey);
	
	std::cout << std::endl;
	std::cout << "Mnemonic: " << mnemonicObj.toStr() << std::endl;
	std::cout << std::endl;

	std::cout << "Priv spend key: " << privSpendKey << std::endl;
	std::cout << "Priv view key : " << privViewKey << std::endl;
	std::cout << std::endl;

	XmrEllipticCurve ec;

	const XmrPubKey pubKeySpend = privSpendKey.getPubkey(ec);
	const XmrPubKey pubKeyView = privViewKey.getPubkey(ec);

	const std::string pubKeySpendStr = pubKeySpend.toString();

	std::cout << "pubKeySpend key: " << reverseIntBytes<256>(pubKeySpend.xyVal()) << std::endl;		
	std::cout << "pubKeyView key : " << reverseIntBytes<256>(pubKeyView.xyVal()) << std::endl;		
	std::cout << std::endl;	

	PubXmrAddress pubXmrAddress(pubKeySpend, pubKeyView);
				
	const std::string convertedAddr = pubXmrAddress.toString();		

	const std::string privHex = hexString(privSpendKey.toBinaryXmrString());

	//const ExtraInfo extraInfoPrivSpend("Priv spend", privSpendKey.toString());
	//privateGenerationExtra.push_back(extraInfoPrivSpend);

	const ExtraInfo extraInfo2("Priv view", privViewKey.toString());
	privateGenerationExtra.push_back(extraInfo2);

	std::cout << "Address: " << convertedAddr << std::endl;
	std::cout << std::endl;		
	
	const GenerationInfoBase infoBase(thetime, getNodeName(), applicationName, versionNumber);
	
	const int newNum = getNextNumberIfConfigured(params);
	
	const GenerationInfoCore infoCore(infoBase, newNum, params);
	GenerationInfo info(infoCore, privateGenerationExtra, genMethod);

	info.setMnemonic(mnemonicObj.toStr());
	info.setInhibitFirstPrivkeyLine(true);

	const AddressGenerationInfo allInfo(pubXmrAddress.toString(), privSpendKey.toString(), "Priv spend", info);
	
	saveAllFiles(allInfo);	
}




bool GenerationMethods::parseXmrCommandlineParams(const std::vector<std::string>& args, const int minNumArgs, const int argStartPos, GenerationParameters& params, ExtraOptionsRandom& extraOptions)
{
	if(args.size() < minNumArgs)
	{
		std::cout << "Not enough arguments" << std::endl;
		return false;
	}
	
	const std::vector<std::string> allExtraArgs = getExtraArgs(args, argStartPos);	
		
	for(std::vector<std::string>::const_iterator it = allExtraArgs.begin() ;
		it != allExtraArgs.end() ;
		++it)
	{
		const std::string& arg = (*it);
		
		if(params.useArg(arg))
		{
			//Nothing
		}
		else if(extraOptions.useArg(arg))
		{
			//Nothing			
		}
		else
		{
			std::cout << "Incorrect argument: " << arg << std::endl;
			throw std::string("Argument error");
			
			//Todo: Return false
		}
	}
	
	params.finally();


	return true;
}


void GenerationMethods::generateMoneroRandom(const std::vector<std::string>& args, const std::string& thetime)
{
	std::cout << "Generating monero private/public key" << std::endl;

	ExtraOptionsRandom extraOptions;
	GenerationParameters params;
		
	if(!parseXmrCommandlineParams(args, 1, 0, params, extraOptions))
	{
		std::cout << "Incorrect usage" << std::endl;
		std::cout << "Usage: bitgen-xmr random [EXTRAARGS]" << std::endl;
		std::cout << "Example: bitgen-xmr random name \"My test\"" << std::endl;
		
	}

	if(extraOptions.useURandom)
	{
		std::cout << "Using /dev/urandom with hash of buffer" << std::endl;		
	}
		
	const std::string generator = extraOptions.useURandom ? "/dev/urandom" : "/dev/random";

	printRandomSourceWarning(generator);
		
	std::cout << "Generating random address" << std::endl;
	if(!extraOptions.useURandom)
	{
		std::cout << "Press random keys or move the mouse if needed" << std::endl;
	}

	const BigInt<256> privKey = extraOptions.useURandom ? getBlockCalcRandomNumber() : getRandomNumber(extraOptions.longbuf);

	std::list<ExtraInfo> privateGenerationExtra;
	
	generateXmrFromRawPrivkey(privKey, params, thetime, generator, privateGenerationExtra);		
}





void GenerationMethods::generateMoneroHash(const std::vector<std::string>& args, const std::string& thetime)
{
	ExtraOptionsRandom extraOptions;
	GenerationParameters params;
	
	if(!parseXmrCommandlineParams(args, 2, 1, params, extraOptions))
	{
		std::cout << "Usage: bitgen hash \"The hashval\" [extraOptions]" << std::endl;
		std::cout << " or" << std::endl;
		std::cout << "       bitgen hash - [extraOptions]" << std::endl;
		
		return;		
	}
	
	
	std::cout << "Generating address from hash" << std::endl;

	std::string hashval(args[1]);
	const int minChars(200);

	
	if(hashval == "-")
	{
		std::cout << "Enter the hash string at one row" << std::endl;
		std::cout << "Press enter when everything is entered" << std::endl;
		
		hashval = readStdinLine(minChars, false, everythingOk);
	}
	else
	{
		std::cout << "WARNING: The information entered at the command line" << std::endl;
		std::cout << "WARNING: might be stored in a shell history file." << std::endl;
		std::cout << "WARNING: Consider using the \"-\" option for entering information." << std::endl;			
		std::cout << std::endl;		
	}

	const BigInt<256> privKeyRaw = hashStringToBigInt(hashval, minChars);

	std::list<ExtraInfo> privateGenerationExtra;

	generateXmrFromRawPrivkey(privKeyRaw, params, thetime, "hash", privateGenerationExtra);	
}




void GenerationMethods::generateMoneroDice(const std::vector<std::string>& args, const std::string& thetime)
{	
	ExtraOptionsRandom extraOptions;
	GenerationParameters params;
	
	if(!parseXmrCommandlineParams(args, 2, 1, params, extraOptions))
	{
		std::cout << "Usage: bitgen dice digits(1-6) [AccountName]" << std::endl;
		std::cout << " or" << std::endl;
		std::cout << "       bitgen dice - [AccountName]" << std::endl;
			
		return;
	}
	
	std::string dice(args[1]);
	
	std::cout << "DICEVAL: " << dice << std::endl;
	
	if(dice == "-")
	{
		std::cout << "Enter the 100 dice values at one row as digits 1-6" << std::endl;
		std::cout << "Press enter when all digits are entered" << std::endl;
		
		dice = readStdinLine(100, true, oneToSix);
	}
	else
	{
		std::cout << "WARNING: The information entered at the command line" << std::endl;
		std::cout << "WARNING: might be stored in a shell history file." << std::endl;
		std::cout << "WARNING: Consider using the \"-\" option for entering information." << std::endl;			
		std::cout << std::endl;
	}
		
	if(dice.size() != 100)
	{
		std::string errMsg("Need exactly 100 dice throws, got: ");
		errMsg += intToString(dice.size());
		errMsg += " throws";
			
		throw errMsg;		
	}

	std::cout << "Generating bitcoin address from dice, got " << dice.size() << " throws" << std::endl;
	std::cout << "Dice result (1-6): " << dice << std::endl;

	std::list<ExtraInfo> privateGenerationExtra;
	const ExtraInfo extraInfo("Dice val", dice);
	
	privateGenerationExtra.push_back(extraInfo);

	const BigInt<256> privkey = getRandomNumFrom6SideDice(dice);		

	generateXmrFromRawPrivkey(privkey, params, thetime, "dice", privateGenerationExtra);
}


void GenerationMethods::generateMoneroPrivSpend(const std::vector<std::string>& args, const std::string& thetime)
{
	ExtraOptionsRandom extraOptions;
	GenerationParameters params;
	
	if(!parseXmrCommandlineParams(args, 2, 1, params, extraOptions))
	{
		std::cout << "Usage: bitgen-" << getCurrencyShortName() << " priv PRIVSPEND [extraOptions]" << std::endl;
		std::cout << " or" << std::endl;
		std::cout << "       bitgen-" << getCurrencyShortName() << " priv - [extraOptions]" << std::endl;
		
		return;			
	}

	std::string privKeyStrMix(args[1]);

	if(privKeyStrMix == "-")
	{
		std::cout << "Enter the 64 hex values at one row" << std::endl;
		std::cout << "Press enter when all digits are entered" << std::endl;
		
		privKeyStrMix = readStdinLine(64, true, hexChar);
	}
	else
	{
		std::cout << "WARNING: The information entered at the command line" << std::endl;
		std::cout << "WARNING: might be stored in a shell history file." << std::endl;
		std::cout << "WARNING: Consider using the \"-\" option for entering information." << std::endl;
		std::cout << std::endl;
	}
	
	
	const BigInt<256> privKey = hexStringToBigInt(privKeyStrMix);

	std::list<ExtraInfo> privateGenerationExtra;
	const ExtraInfo extraInfo("Spend privkey", hexString(privKey.getBinaryLong()));
	privateGenerationExtra.push_back(extraInfo);				

	generateXmrFromPrivSpendKey(privKey, params, thetime, "spend privkey", privateGenerationExtra);
	
}


void GenerationMethods::generateMoneroHex(const std::vector<std::string>& args, const std::string& thetime)
{
	ExtraOptionsRandom extraOptions;
	GenerationParameters params;
	
	if(!parseXmrCommandlineParams(args, 2, 1, params, extraOptions))
	{
		std::cout << "Usage: bitgen-" << getCurrencyShortName() << " hex hexdigits [extraOptions]" << std::endl;
		std::cout << " or" << std::endl;
		std::cout << "       bitgen-" << getCurrencyShortName() << " hex - [extraOptions]" << std::endl;
		
		return;			
	}

	std::string privKeyStrMix(args[1]);

	if(privKeyStrMix == "-")
	{
		std::cout << "Enter the 64 hex values at one row" << std::endl;
		std::cout << "Press enter when all digits are entered" << std::endl;
		
		privKeyStrMix = readStdinLine(64, true, hexChar);
	}
	else
	{
		std::cout << "WARNING: The information entered at the command line" << std::endl;
		std::cout << "WARNING: might be stored in a shell history file." << std::endl;
		std::cout << "WARNING: Consider using the \"-\" option for entering information." << std::endl;
		std::cout << std::endl;
	}
	
	
	const BigInt<256> privKey = hexStringToBigInt(privKeyStrMix);

	std::list<ExtraInfo> privateGenerationExtra;
	const ExtraInfo extraInfo("Hex privkey", hexString(privKey.getBinaryLong()));
	privateGenerationExtra.push_back(extraInfo);				

	generateXmrFromRawPrivkey(privKey, params, thetime, "hex input", privateGenerationExtra);
}







void GenerationMethods::generateMoneroFromMnemonic(const std::vector<std::string>& args, const std::string& thetime)
{
	ExtraOptionsRandom extraOptions;
	GenerationParameters params;
	
	if(!parseXmrCommandlineParams(args, 2, 1, params, extraOptions))
	{	
		std::cout << "Usage: bitgen-xmr mnemonic MNEMONIC [extraOptions]" << std::endl;
		return;					
	}
	
	std::string mnemonicStr(args[1]);
			
	const XmrMnemonic mnemonicObj(mnemonicStr);
	const BigInt<256> privKeyFinal = mnemonicObj.toPrivKeyInt();
	
	std::list<ExtraInfo> privateGenerationExtra;
	
	generateXmrFromRawPrivkey(privKeyFinal, params, thetime, "mnemonic input", privateGenerationExtra);
}



void GenerationMethods::generateMoneroBrainWallet(const std::vector<std::string>& args, const std::string& thetime)
{	
#ifdef NO_BRAINWALLET_GENERATION
		std::cout << "No brainwallet support" << std::endl;
		return;
#else		
	
	ExtraOptionsRandom extraOptions;
	GenerationParameters params;
	
	if(!parseXmrCommandlineParams(args, 3, 2, params, extraOptions))
	{	
		std::cout << "Usage: bitgen brainwallet email/salt \"The passphrase\" [extraOptions]" << std::endl;
		return;
	}
	
	const std::string salt(args[1]);

	std::string passphrase(args[2]);
	if(passphrase == "-")
	{
		std::cout << "Enter the brainwallet phrase" << std::endl;
		std::cout << "Press enter when all characters are entered" << std::endl;
		
		passphrase = readStdinLine(50, false, everythingOk);
	}
	else
	{
		std::cout << "WARNING: The information entered at the command line" << std::endl;
		std::cout << "WARNING: might be stored in a shell history file." << std::endl;
		std::cout << "WARNING: Consider using the \"-\" option for entering information." << std::endl;			
		std::cout << std::endl;		
	}

	std::cout << "Generating brainwallet" << std::endl;

	std::cout << "Salt: \"" << salt << "\"" << std::endl;		
	std::cout << "Passphrase: \"" << passphrase << "\"" << std::endl;


	const BigInt<256> privKey =  brainwalletToBigInt(passphrase, salt);
		
	std::list<ExtraInfo> privateGenerationExtra;
	const ExtraInfo extraInfo("Passphrase", passphrase);
	privateGenerationExtra.push_back(extraInfo);		

	
	generateXmrFromRawPrivkey(privKey, params, thetime, "brainwallet", privateGenerationExtra);

#endif
		
}





void GenerationMethods::generateMoneroOtpParts(const std::vector<std::string>& args, const std::string& thetime)
{
	if(args.size() < 3)
	{
		std::cout << "Usage: bitgen otpgen [NUMPAD/useconf] METHOD [extraOptions]" << std::endl;
		std::cout << "  Methods: random, dice, hash, hex" << std::endl;
		
		throw std::string("Incorrect usage");
	}
	
	const std::string numStr = args[1];
	std::list<std::string> folders;
	
	int totalPads(0);
	if(numStr == "useconf")
	{
		std::cout << "Using configured folders for num pads" << std::endl;
		std::string publicFolder;
		folders = BitgenConfig::getPublicAndOtpFolders(publicFolder);
		
		BitgenConfig::checkFoldersExist(folders);
		std::cout << "OK, configured folders exist" << std::endl;
		
		totalPads = folders.size();
	}
	else
	{
		totalPads = atoi(numStr.c_str());
		
		for(int i = 0 ; i < totalPads ; i++)
		{
			folders.push_back(".");
		}
	}
	
	
	
	if(totalPads < 2)
	{
		std::cout << "Num pads given: " << totalPads << " (" << numStr << ")" << std::endl;
		std::cout << "Num pads must be at least 2" << std::endl;
		throw std::string("Incorrect number of pads");
	}
	
	std::cout << "Total pads: " << totalPads << std::endl;

	const std::string method = args[2];
	
	if(method != "random" && method != "hash" && method != "dice" && method != "hex")
	{
		std::cout << "Unknown method: " << method << std::endl;
		throw std::string("Unknown method");
	}
	
	std::string extraVal;


	if(method == "random")
	{
		//Nothing
	}
	else if(method == "hash")
	{
		if(args.size() < 4)
		{
			std::cout << "Usage: bitgen otpgen NUMPAD hash HASHVAL [extraOptions]" << std::endl;
		
			throw std::string("Incorrect usage");
		}
		
		extraVal = args[3];		//Todo: Rename to paramVal
	}
	else if(method == "dice")
	{
		if(args.size() < 4)
		{
			std::cout << "Usage: bitgen otpgen NUMPAD dice DICEVAL [extraOptions]" << std::endl;
		
			throw std::string("Incorrect usage");
		}
		
		extraVal = args[3];		//Todo: Rename to paramVal
	}
	else if(method == "hex")
	{
		if(args.size() < 4)
		{
			std::cout << "Usage: bitgen otpgen NUMPAD hex HEXVAL [extraOptions]" << std::endl;
		
			throw std::string("Incorrect usage");
		}
		
		extraVal = args[3];		//Todo: Rename to paramVal
	}
	else
	{
		std::cout << "Internal error, no such method: " << method << std::endl;
		throw std::string("No such method");
	}


	const int paramStartPos = (method == "random") ? 2 : 3;


	ExtraOptionsRandom extraOptions;
	GenerationParameters params;
	
	if(!parseXmrCommandlineParams(args, 3, paramStartPos, params, extraOptions))
	{	
			std::cout << "Incorrect arguments " << std::endl;		
			throw std::string("Argument error");			
	}
	
	bool useFolder = params.getUseFolder();
	bool useConfiguredFolderPaths(false);
	
	if(numStr == "useconf")		//Using folder conf implies that folders are used
	{
		useFolder = true;
		useConfiguredFolderPaths = true;
	}

	const std::list<BigInt<256> > pads = generatePadRandom(totalPads, method, extraVal);


	std::cout << std::endl;

		
	BigInt<256> privKey;
	for(std::list<BigInt<256> >::const_iterator it = pads.begin() ;
		it != pads.end() ;
		++it)
	{
		const BigInt<256>& pad = (*it);
		std::cout << "Pad: " << hexString(pad.getBinaryLong()) << std::endl;
		
		privKey = privKey ^ pad;
	}

	//Check if a static pad is defined
	const BigInt<256> staticXor = BitgenConfig::getStaticXor();
	//std::cout << "Static xor: " << hexString(staticXor.getBinaryLong()) << std::endl;
	
	const BigInt<256> zero;
	
	const bool useStaticPad = (staticXor != zero);
	
	std::list<ExtraInfo> generationExtra;	
	
	if(useStaticPad)
	{
		const std::string hash = staticXorHash(staticXor);		
		
		std::cout << "Using static pad: " <<  hexString(staticXor.getBinaryLong()) << " (" <<
			hexString(hash) << ")" << std::endl;
			
		const ExtraInfo extraInfoStaticPad("Static pad", hexString(hash));	
		generationExtra.push_back(extraInfoStaticPad);			
	}

	privKey = privKey ^ staticXor;
	
	//limit to max value
	const BigInt<256> finalPrivInt = XmrEllipticCurve::getEdwards25519Max(privKey);
	
	std::cout << "Privkey: " << hexString(finalPrivInt.getBinaryLong()) << std::endl;
		
	const XmrPrivSpendKey myPrivKeySpend(finalPrivInt);
	const XmrPrivViewKey myViewKey = myPrivKeySpend.privateSpendToView();
	
	const XmrMnemonic mnemonicObj(myPrivKeySpend);
	
	std::cout << std::endl;
	std::cout << "Mnemonic: " << mnemonicObj.toStr() << std::endl;
	std::cout << std::endl;

	std::cout << "Priv spend key: " << myPrivKeySpend << std::endl;
	std::cout << "Priv view key : " << myViewKey << std::endl;
	std::cout << std::endl;

	XmrEllipticCurve ec;

	const XmrPubKey pubKeySpend = myPrivKeySpend.getPubkey(ec);
	const XmrPubKey pubKeyView = myViewKey.getPubkey(ec);

	const std::string pubKeySpendStr = pubKeySpend.toString();

	std::cout << "pubKeySpend: " <<  pubKeySpendStr << std::endl;		
	std::cout << "pubKeyView : " << pubKeyView.toString() << std::endl;		
	std::cout << std::endl;

	const PubXmrAddress pubXmrAddress(pubKeySpend, pubKeyView);
				
	const std::string xmrAddr = pubXmrAddress.toString();		

	const std::string generator = ((method == "random") ? "OTP /dev/random" : ("OTP " + method));
	
	std::cout << std::endl;

	std::vector<std::string> privatePaths;	
	std::string pubPath;
	
	std::cout << "Address: " << xmrAddr << std::endl;
	
	
	const int newNum = SaveFolderHandler::getNextNumber();
	//std::cout << "The new number: " << newNum << std::endl;
	//params.setAccountNameForNumber(true, newNum);
	
	if(useFolder)
	{		
		//params.setAccountNameforFolder(true, newNum);
		
		//std::cout << "Account name: " << params.getAccountName() << std::endl;

		const std::string firstFolder = folders.front();

		const std::pair<std::string, std::vector<std::string> > fullFolders = 
			SaveFolderHandler::createNewFolders(newNum, folders, true, useConfiguredFolderPaths);
		
		privatePaths = fullFolders.second;
		pubPath = fullFolders.first;
	}
	else
	{
		for(int i = 0 ; i < totalPads ; i++)
		{
			privatePaths.push_back(".");
		}
		
		pubPath = ".";
	}

	std::cout << std::endl;
	

	const std::list<std::string> hashParts = splitHashStr(totalPads, extraVal);

	const GenerationInfoBase infoBase(thetime, getNodeName(), applicationName, versionNumber);
	const GenerationInfoCore infoCore(infoBase, newNum, params);
	const GenerationInfo info(infoCore, generationExtra, generator);	//"otp"


	savePrivatePadPart(info, pads, hashParts, privatePaths, totalPads);
		//pads, method, hashParts, privatePaths, xmrAddr,
		//generator, thetime, totalPads, params, staticXor, newNum);

	std::cout << std::endl;

	const AddressGenerationInfo allInfo(xmrAddr, "DummyPrivAddr", "Dummy", info);


	savePublicPadPart(allInfo, pubPath);

	if(useStaticPad)
	{
		std::cout << std::endl;
		std::cout << "NOTE: STATIC PAD USED. THE PAD MUST BE SECURELY STORED" << std::endl;
		std::cout << "Without knowledge of the static pad, the private key can not be recovered" << std::endl;
	}	
}



void GenerationMethods::handleOtpPrivKey(const GenerationParameters& params, const BigInt<256>& staticXor, const int numPads, const BigInt<256>& finalPrivInt, const std::string& thetime)
{	
	std::cout << std::endl;
	std::cout << "Privkey: " << hexString(finalPrivInt.getBinaryLong()) << std::endl;

	const XmrPrivSpendKey myPrivKeySpend(finalPrivInt);
	const XmrPrivViewKey myViewKey = myPrivKeySpend.privateSpendToView();
	
	const XmrMnemonic mnemonicObj(myPrivKeySpend);
	
	std::cout << std::endl;
	std::cout << "Mnemonic: " << mnemonicObj.toStr() << std::endl;
	std::cout << std::endl;

	//std::cout << "Priv spend key: " << myPrivKeySpend << std::endl;
	//std::cout << "Priv view key : " << myViewKey << std::endl;
	//std::cout << std::endl;

	XmrEllipticCurve ec;

	const XmrPubKey pubKeySpend = myPrivKeySpend.getPubkey(ec);
	const XmrPubKey pubKeyView = myViewKey.getPubkey(ec);

	const std::string pubKeySpendStr = pubKeySpend.toString();

	//std::cout << "pubKeySpend: " <<  pubKeySpendStr << std::endl;		
	//std::cout << "pubKeyView : " << pubKeyView.toString() << std::endl;		
	//std::cout << std::endl;

	PubXmrAddress pubXmrAddress(pubKeySpend, pubKeyView);

	std::cout << "Address: " << pubXmrAddress.toString() << std::endl;
	std::cout << std::endl;

	const std::string hexStr = hexString(finalPrivInt.getBinaryLong());

	std::list<ExtraInfo> generationExtra;
	
	//const ExtraInfo extraInfo("SpendKey", myPrivKeySpend.toString());
	//generationExtra.push_back(extraInfo);

	const ExtraInfo extraInfoPrivView("Priv Viewkey", myViewKey.toString());
	generationExtra.push_back(extraInfoPrivView);

	
	//const ExtraInfo extraInfoParts("OTP parts", intToString(numPads));
	//generationExtra.push_back(extraInfoParts);

	const std::string generator = "OTP with " + intToString(numPads) + " parts";	
	const int newNum(0);

	const GenerationInfoBase infoBase(thetime, getNodeName(), applicationName, versionNumber);
	const GenerationInfoCore infoCore(infoBase, newNum, params);
	GenerationInfo info(infoCore, generationExtra, generator);

	info.setMnemonic(mnemonicObj.toStr());
	info.setIsPad(false);
	info.setStaticXor(staticXor);

	const AddressGenerationInfo allInfo(pubXmrAddress.toString(), myPrivKeySpend.toString(), "Priv spend", info);

	saveAllFiles(allInfo);
}



void GenerationMethods::generateMoneroOtpTotal(const std::vector<std::string>& args, const std::string& thetime)
{
	std::cout << "Generating OTP total" << std::endl;
	
	if(args.size() < 3)
	{
		std::cout << "Usage: bitgen otp OTP1 ... OTPn" << std::endl;
		return;
	}
	
	//Todo: Extra options
	
	GenerationParameters params;	
	params.setAccountName("From OTP do not use");


	bool useStaticXorIfPresent(true);
	std::list<BigInt<256> > padList = getPadsFromCommandline(args, useStaticXorIfPresent);

	const int numPads = padList.size();
	
	std::cout << "Num pads: " << numPads << std::endl;

	const BigInt<256> staticXor = BitgenConfig::getStaticXor(useStaticXorIfPresent);
	
	const BigInt<256> zero;
	if(staticXor != zero)
	{
		const std::string hash = staticXorHash(staticXor);		
		std::cout << "Using static pad: " <<  hexString(staticXor.getBinaryLong()) << " ";
		std::cout << "(" << hexString(hash) << ")" << std::endl;	
	}
	

	BigInt<256> privKey;
	
	for(std::list<BigInt<256> >::const_iterator it = padList.begin() ;
		it != padList.end() ;
		++it)
	{		
		const BigInt<256> padInt = (*it);
		
		std::cout << "PAD: " << hexString(padInt.getBinaryLong()) << std::endl;

		privKey = padInt ^ privKey;
	}
		
	privKey = staticXor ^ privKey;

	//limit to max value
	const BigInt<256> finalPrivInt = XmrEllipticCurve::getEdwards25519Max(privKey);

	handleOtpPrivKey(params, staticXor, numPads, finalPrivInt, thetime);
}









void GenerationMethods::infoAddress(const std::vector<std::string>& args, const std::string& thetime)
{

	std::cout << "Address information" << std::endl;

	if(args.size() < 2)
	{
		std::cout << "Usage: bitgen-xmr addr ADDR  [extraOptions]" << std::endl;
		return;					
	}
	
	const std::string pubAddr = args[1];
	std::cout << "Pubaddr: " << pubAddr << std::endl;
				
	const PubXmrAddress pubXmrAddress(pubAddr);
		
	const BigInt<257> pubSpendKey = pubXmrAddress.getPubSpend();
	const BigInt<257> pubViewkey = pubXmrAddress.getPubView();
	
	std::cout << "pubSpend key: " << reverseIntBytes<257>(pubSpendKey) << std::endl;
	std::cout << "pubView key : " << reverseIntBytes<257>(pubViewkey) << std::endl;
	std::cout << std::endl;


	std::cout << "pubSpend point: " << pubXmrAddress.getSpendKey().point.toString() << std::endl;
	std::cout << "pubView point : " << pubXmrAddress.getViewKey().point.toString() << std::endl;

		
	//const std::string pubSpendStr = pubSpend.getBinary();
	//const std::string pubViewStr = pubView.getBinary();
	
	const std::string convertedAddr = pubXmrAddress.toString();		
		
	if(pubAddr != convertedAddr)
	{
		std::cout << "pubAddr      : " << pubAddr << std::endl;
		std::cout << "convertedAddr: " << convertedAddr << std::endl;
		
		throw std::string("Error, addresses does not match");
	}

	const std::string xmrPubAddress = convertedAddr;

}




void infoAboutXmrEntity(const std::vector<std::string>& args, const std::string& thetime)
{
	if(args.size() < 2)
	{
		std::cout << "Usage: bitgen-xmr info address" << std::endl;
		return;			
	}

		
	const std::string adr(args[1]);
	const PubXmrAddress pubXmrAddress(adr);
	std::cout << "Address     : " << pubXmrAddress.toString() << std::endl;
	std::cout << std::endl;

	if(pubXmrAddress.hasIdPayment())
	{
		std::cout << "Payment id  : " << hexString(pubXmrAddress.getPaymentId().getBinaryShortest()) << std::endl;
	}
	
	if(pubXmrAddress.isSubAddr())
	{
		std::cout << "Is subaddress" << std::endl;
	}
	else
	{
		std::cout << "Is NOT subaddress" << std::endl;
	}
		
		
	std::cout << "Public spend: " << hexString(reverseIntBytes<257>(pubXmrAddress.getPubSpend()).getBinaryShortest()) << std::endl;
	std::cout << "Public view : " << hexString(reverseIntBytes<257>(pubXmrAddress.getPubView()).getBinaryShortest()) << std::endl;	


	const PubXmrAddress addressWithputPaymentId(pubXmrAddress, false);
	std::cout << "Base address: " << addressWithputPaymentId.toString() << std::endl;


	std::cout << std::endl;
}






void addPaymentId(const std::vector<std::string>& args, const std::string& thetime)
{
	if(args.size() < 2)
	{
		std::cout << "Usage: bitgen-xmr paymentId address [paymentId]" << std::endl;
		std::cout << "       paymentId is a 16 digit hex string (64 bits)" << std::endl;		
		std::cout << "       Without payment id a random number will be generated" << std::endl;				
		return;			
	}
		
	std::string payementIdStr;
	if(args.size() > 2)
	{
		payementIdStr = args[2];
		//std::cout << "payementIdStr  : " << payementIdStr << std::endl;
		
	}
		
		
	const std::string adr(args[1]);
	const PubXmrAddress pubXmrAddress(adr);

	if(pubXmrAddress.isSubAddr())
	{
		std::string errStr("Can not add payment id to subaddress: ");
		errStr += pubXmrAddress.toString();
		
		throw errStr;
	}

	std::cout << "Original address   : " << pubXmrAddress.toString() << std::endl;


	if(pubXmrAddress.hasIdPayment())
	{
		std::cout << "Existing payment id: " << hexString(pubXmrAddress.getPaymentId().getBinaryShortest()) << std::endl;
	}


	std::cout << std::endl;


	std::string random;

	if(payementIdStr == "")
	{
		std::cout << "Generating random payment id. Generate entropy if needed, for example by moving the mouse." << std::endl;
		random = getFromDevRandom(8);
	}
	
	const BigInt<65> paymentId = (payementIdStr == "") ?  BigInt<65>(random, 'c') : BigInt<65>(payementIdStr.c_str());
	
	std::cout << "Setting payment id : " << hexString(paymentId.getBinaryShortest()) << std::endl;

	const PubXmrAddress addressWithPaymentId(pubXmrAddress, paymentId);
	
	std::cout << "Integrated address : " << addressWithPaymentId.toString() << std::endl;	
}





void generateMoneroPaymentRequest(const std::vector<std::string>& args, const std::string& thetime)
{
	if(args.size() < 3)
	{
		std::cout << "Usage: bitgen-xmr invoice coinAddress amount [recipient name/label] [message] [LastDate]" << std::endl;
		return;			
	}
		
	const std::string adr(args[1]);

	const PubXmrAddress pubXmrAddress(adr);
	
	std::cout << "Address: " << pubXmrAddress.toString() << std::endl;


	std::string amount;
	if(args.size() > 2)
	{
		const std::string amountGiven(args[2]); 
		//Make sure the value is a non-zero number
		const double value = atof(amountGiven.c_str());
		if(value == 0.0)
		{
			throw std::string("Incorrect or zero sum given");
		}
		
		std::cout << "Amount : " << value << std::endl;
		char str[128];
		sprintf(str, "%0f", value);
		amount = str;
	}


	std::string label;
	if(args.size() > 3)
	{
		label = args[3];
		std::cout << "Label  : " << label << std::endl;
		
	}

	std::string message;
	if(args.size() > 4)
	{
		message = args[4];
		std::cout << "Message: " << message << std::endl;		
	}

	std::string lastDate;
	if(args.size() > 5)
	{
		lastDate = args[5];		
		std::cout << "Date   : " << lastDate << std::endl;		
	}
	
	
	const std::list<std::string> not_used_printExtra;

	const Language lang = getLanguage();


	const GenerationInfoBase infoBase(thetime, getNodeName(), applicationName, versionNumber);

	std::list<ExtraInfo> generationExtra;
	GenerationParameters params;
	params.setAccountName(label);

	const int newNum = GenerationMethods::getNextNumberIfConfigured(params);

	
	const GenerationInfoCore infoCore(infoBase, newNum, params);
	const GenerationInfo info(infoCore, generationExtra, "zzz");
	const AddressGenerationInfo addrInfo(pubXmrAddress.toString(), "", "", info);


	const PostscriptGenerator postscriptGenerator(info, versionNumber, applicationName, lang);		
	const PublicPostscriptInfo pubInfo(addrInfo);
	
	const std::string paymentId = pubXmrAddress.hasIdPayment() ? hexString(pubXmrAddress.getPaymentId().getBinaryShortest()) : "";
	
	const std::string psPublic = postscriptGenerator.generateInvoice(pubInfo, amount, message, lastDate, paymentId);

	std::cout << std::endl;
	
	const std::string pubFilename = "num_" + intToString(info.getNewNum()) + "_" + pubInfo.pubkey.substr(0, 8) + "_invoice.ps";
	std::cout << "Saving publfile: " << pubFilename << std::endl;
	fileWrite(pubFilename, psPublic);	
}





void generateMoneroAddressParts(const std::vector<std::string>& args, const std::string& thetime)
{
	if(args.size() < 2)
	{
		std::cout << "Usage: bitgen-xmr pub coinAddress [message] [LastDate]" << std::endl;
		return;			
	}
		
	const std::string adr(args[1]);

	const PubXmrAddress pubXmrAddress(adr);
	
	std::cout << "Address: " << pubXmrAddress.toString() << std::endl;

	std::string label;
	if(args.size() > 2)
	{
		label = args[2];
		std::cout << "Label  : " << label << std::endl;
		
	}
	

	const Language lang = getLanguage();


	const GenerationInfoBase infoBase(thetime, getNodeName(), applicationName, versionNumber);
	std::list<ExtraInfo> generationExtra;
	const GenerationParameters params;

	const int newNum = GenerationMethods::getNextNumberIfConfigured(params);


	const GenerationInfoCore infoCore(infoBase, newNum, params);
	const GenerationInfo info(infoCore, generationExtra, "zzz");
	const AddressGenerationInfo addrInfo(pubXmrAddress.toString(), "", "", info);


	const PostscriptGenerator postscriptGenerator(info, versionNumber, applicationName, lang);		
	const PublicPostscriptInfo pubInfo(addrInfo);
	
	const std::string psPublic = postscriptGenerator.postscriptPagePublic(pubInfo);

	std::cout << std::endl;

	const std::string pubFilename = "xmr_" + intToString(info.getNewNum()) + "_" + pubInfo.pubkey.substr(0, 8) + "_share.ps";
	std::cout << "Saving publfile: " << pubFilename << std::endl;
	fileWrite(pubFilename, psPublic);
}




void GenerationMethods::deriveMoneroAddressesFromPrivView(const std::vector<std::string>& args, const std::string& thetime)
{
	std::cout << "Generating subaddresses from priv view key" << std::endl;

	ExtraOptionsRandom extraOptions;
	GenerationParameters params;
	
	if(!parseXmrCommandlineParams(args, 2, 4, params, extraOptions))
	{	
		std::cout << "Usage: bitgen-xmr deriveprivview ADDRESS PRIVVIEW [numSubaddr] [accountNr]" << std::endl;
		return;					
	}


	const std::string adr(args[1]);
	//std::cout << "The addr: " << adr << std::endl;

	const PubXmrAddress baseAddress(adr);

	
	std::string privViewStr(args[2]);
	
	int numSubAddr(10);
	if(args.size() > 3)
	{
		const std::string numSubAddrStr = args[3];
		numSubAddr = stoi(numSubAddrStr);
		std::cout << "Number of subaddr: " << numSubAddr << std::endl;
	}	

	int account(0);
	if(args.size() > 4)
	{
		const std::string accountNumStr = args[4];
		account = stoi(accountNumStr);
		//std::cout << "Account num: " << account << std::endl;
	}	

	

	const BigInt<257> privViewInt(privViewStr.c_str());
	const BigInt<257> privViewReversed = reverseIntBytes<257>(privViewInt);	
	const BigInt<256> privViewInt256(privViewReversed);
	
	const XmrPrivViewKey privViewKey(privViewInt256);
	
	const XmrEllipticCurve ec;


	std::cout << "Account: " << std::dec << account << std::endl;
	for(int index = 0 ; index < numSubAddr ; index++)
	{
		const PubXmrAddress pubXmrAddress = privViewKey.getSubAddressFromViewKey(baseAddress.getSpendKey(), ec, account, index);

		std::cout << "" << std::dec << std::setfill(' ') << std::setw(3) << index << ": " << pubXmrAddress.toString() << std::endl;
	}
	std::cout << std::endl;
}






void GenerationMethods::deriveMoneroAddressesFromPrivSpend(const std::vector<std::string>& args, const std::string& thetime)
{
	std::cout << "Generating subaddresses from priv spend key" << std::endl;

	ExtraOptionsRandom extraOptions;
	GenerationParameters params;
	
	if(!parseXmrCommandlineParams(args, 2, 3, params, extraOptions))
	{	
		std::cout << "Usage: bitgen-xmr deriveprivspend PRIVSPEND [numSubaddr] [accountNr]" << std::endl;
		return;					
	}
	
	std::string privSpendStr(args[1]);
	
	int numSubAddr(10);
	if(args.size() > 2)
	{
		const std::string numSubAddrStr = args[2];
		numSubAddr = stoi(numSubAddrStr);
		std::cout << "Number of subaddr: " << numSubAddr << std::endl;
	}	

	int account(0);
	if(args.size() > 3)
	{
		const std::string accountNumStr = args[3];
		account = stoi(accountNumStr);
		std::cout << "Account num: " << account << std::endl;
	}	

	const BigInt<257> privSpendInt(privSpendStr.c_str());
	//std::cout << "privSpendInt: " << privSpendInt << std::endl;

	const BigInt<257> privSpendReversed = reverseIntBytes<257>(privSpendInt);
	
	const BigInt<256> privSpendInt256(privSpendReversed);
	//std::cout << "privSpendInt2: " << privSpendInt2 << std::endl;
	
	const XmrPrivSpendKey privSpendKey(privSpendInt256); //, true);
	std::cout << "privSpendKey: " << privSpendKey << std::endl;
	
	const XmrEllipticCurve ec;

	std::cout << "Account: " << std::dec << account << std::endl;
	for(int index = 0 ; index < numSubAddr ; index++)
	{
		const PubXmrAddress pubXmrAddress = privSpendKey.getSubAddress(ec, account, index);

		std::cout << "" << std::dec << std::setfill(' ') << std::setw(3) << index << ": " << pubXmrAddress.toString() << std::endl;
	}
	std::cout << std::endl;
}





void GenerationMethods::deriveMoneroAddresses(const std::vector<std::string>& args, const std::string& thetime)
{
	std::cout << "Generating subaddresses" << std::endl;

	ExtraOptionsRandom extraOptions;
	GenerationParameters params;
	
	if(!parseXmrCommandlineParams(args, 2, 3, params, extraOptions))
	{	
		std::cout << "Usage: bitgen-xmr derive MNEMONIC [numSubaddr] [accountNr]" << std::endl;
		return;					
	}
	
	std::string mnemonicStr(args[1]);
	
	int numSubAddr(10);
	if(args.size() > 2)
	{
		const std::string numSubAddrStr = args[2];
		numSubAddr = stoi(numSubAddrStr);
		std::cout << "Number of subaddr: " << numSubAddr << std::endl;
	}	

	int account(0);
	if(args.size() > 3)
	{
		const std::string accountNumStr = args[3];
		account = stoi(accountNumStr);
		std::cout << "Account num: " << account << std::endl;
	}	

			
	const XmrMnemonic mnemonicObj(mnemonicStr);
	std::cout << "Mnemonic: " << mnemonicObj.toStr() << std::endl;
	
	const XmrPrivSpendKey privSpendKey = mnemonicObj.getSpendKey();
	
	const XmrEllipticCurve ec;

	std::cout << "Account: " << std::dec << account << std::endl;
	for(int index = 0 ; index < numSubAddr ; index++)
	{
		const PubXmrAddress pubXmrAddress = privSpendKey.getSubAddress(ec, account, index);

		std::cout << "" << std::dec << std::setfill(' ') << std::setw(3) << index << ": " << pubXmrAddress.toString() << std::endl;
	}
	std::cout << std::endl;
}


void printXmrUsage()
{
	//std::cout << "bitgen-xmr usage:" << std::endl;

	std::cout << std::endl;		
	std::cout << getCurrencyName() << " address generator" << std::endl;
	std::cout << "Usage: bitgen-" << getCurrencyShortName() << " command [options]" << std::endl;
		
	std::cout << "Key generation: " << std::endl;
	std::cout << "mnemonic | random | hash | dice | hex | brainwallet | priv" << std::endl;
	std::cout << std::endl;

	std::cout << "Integrated address generation: " << std::endl;
	std::cout << "paymentId" << std::endl;
	std::cout << std::endl;

	std::cout << "Subaddress derivation: " << std::endl;
	std::cout << "derive | deriveprivspend | deriveprivview" << std::endl;
	std::cout << std::endl;

	std::cout << "Public address postscript and invoice generation: " << std::endl;
	std::cout << "invoice | pub" << std::endl;
	std::cout << std::endl;

	std::cout << "One-time-pad: " << std::endl;
	std::cout << "otpgen | otp | otpdir | removeotpdir" << std::endl;
	std::cout << std::endl;

	std::cout << "Address numbering configuration: " << std::endl;
	std::cout << "getnextnum | setnextnum" << std::endl;
	std::cout << std::endl;
}


void doMoneroWork(int argc, char* argv[])
{			
	time_t t = time(0);
	const std::string thetime = removeNewline(asctime(localtime(&t)));
	
	std::vector<std::string> args;
	for(int i = 1 ; i < argc ; i++)
	{
		args.push_back(argv[i]);		
	}

	if(argc < 2)
	{
		std::cout << "For help, use \"bitgen-" << getCurrencyShortName() << " help\"" << std::endl;
		throw std::string("Missing argument");
		return;
	}

	const std::string firstArg(argv[1]);


	if(firstArg == "version" || firstArg == "-v")
	{
		printCompleteVersion();
	}
	else if(firstArg == "mnemonic")
	{
		std::cout << "Will generate from mnemonic" << std::endl;
		GenerationMethods::generateMoneroFromMnemonic(args, thetime);		
	}		
	else if(firstArg == "random")
	{
		GenerationMethods::generateMoneroRandom(args, thetime);
	}
	else if(firstArg == "hash")
	{
		GenerationMethods::generateMoneroHash(args, thetime);
	}
	else if(firstArg == "dice")
	{
		GenerationMethods::generateMoneroDice(args, thetime);
	}	
	else if(firstArg == "hex")
	{
		GenerationMethods::generateMoneroHex(args, thetime);
	}
	else if(firstArg == "priv")
	{
		GenerationMethods::generateMoneroPrivSpend(args, thetime);
	}		
	else if(firstArg == "brainwallet")
	{
		GenerationMethods::generateMoneroBrainWallet(args, thetime);
	}
	else if(firstArg == "otpgen")
	{
		GenerationMethods::generateMoneroOtpParts(args, thetime);
	}
	else if(firstArg == "otp")
	{
		GenerationMethods::generateMoneroOtpTotal(args, thetime);
	}
	else if(firstArg == "otpdir")
	{
		setOtpFolders(args);
	}
	else if(firstArg == "removeotpdir")
	{
		removeOtpFolders(args);
	}
	else if(firstArg == "otplist")
	{
		GenerationMethods::listOtpFolder(args, thetime);
	}
	else if(firstArg == "otpempty")
	{
		GenerationMethods::markOtpAsEmptyAmount(args, thetime);
	}	
	else if(firstArg == "addr")
	{
		GenerationMethods::infoAddress(args, thetime);
	}
	else if(firstArg == "invoice")
	{
		generateMoneroPaymentRequest(args, thetime);
	}
	else if(firstArg == "pub")
	{
		generateMoneroAddressParts(args, thetime);
	}	
	else if(firstArg == "derive")
	{
		GenerationMethods::deriveMoneroAddresses(args, thetime);
	}
	else if(firstArg == "deriveprivspend")
	{
		GenerationMethods::deriveMoneroAddressesFromPrivSpend(args, thetime);
	}
	else if(firstArg == "deriveprivview")
	{
		GenerationMethods::deriveMoneroAddressesFromPrivView(args, thetime);
	}	
	else if(firstArg == "info")
	{
		infoAboutXmrEntity(args, thetime);
	}	
	else if(firstArg == "help")
	{
		printXmrUsage();	
	}
	else if(firstArg == "license")
	{
		printLicense();
	}
	else if(firstArg == "getnextnum")
	{
		std::cout << "Getting next num" << std::endl;
		getNextNumToUse();
	}
	else if(firstArg == "setnextnum")
	{		
		std::cout << "Setting next num" << std::endl;
		setNextNumToUse(args);
	}	
	else if(firstArg == "paymentId")
	{		
		std::cout << "Setting payment id" << std::endl;
		addPaymentId(args, thetime);
	}		
	else
	{
		std::cerr << "Unknown command: " << argv[1] << std::endl;
	}
	
}








int main(int argc, char* argv[])
{				
	try
	{
		if(argc > 1)
		{		
		const std::string firstArg(argv[1]);
		if(firstArg == "test")
		{
			unitTestsMonero();
		
			std::cout << std::endl;
			std::cout << "All tests performed successfully" << std::endl;	
			return 0;
		}
		}
		
		doMoneroWork(argc, argv);
	}
	catch(const std::string& err)
	{
		std::cerr << "Error: " << err << std::endl;
		return 1;
	}
	
	return 0;
}


	
/*
	const std::string privkeyRaw = myPrivKeySpend.toXmrString();
	std::cout << "+++privkeyRaw      : " << hexString(privkeyRaw) << std::endl;


	//reverseOrder
	const std::string toHash = privkeyRaw; //subAddr + privkeyRaw + accountIndexStr + subaddressIndexWithinAccountStr;

	const std::string hashViewKey = computeKeccak1600(toHash);

	


	//std::cout << std::endl;
	//std::cout << "+++hashViewKey: " << hexString(hashViewKey) << std::endl;

	const std::string hashViewKeyReversed = reverseOrder(hashViewKey);
	//std::cout << "+++hashRed: " << hexString(hashViewKeyReversed) << std::endl;


	const BigInt<256> hashAsInt(BigInt<256>::fromBytes(hashViewKeyReversed));
	//std::cout << "+++hashAsInt  : " << hashAsInt << std::endl;

	BigInt<256> finalVal = getEdwards25519Max(hashAsInt);
	//std::cout << "+++finalVal   : " << finalVal << std::endl;

	const std::string finalValStr = finalVal.getBinaryLong();
	//std::cout << "+++finalValStr   : " << hexString(finalValStr) << std::endl;

	const std::string finalValStrRev = reverseOrder(finalValStr);
	std::cout << "+++finalValStrRev   : " << hexString(finalValStrRev) << std::endl;


	BigInt<256> privateViewKey(BigInt<256>::fromBytes(finalValStr));

	std::cout << "+++privateViewKey   : " << privateViewKey << std::endl;


	const PrivKey myViewKey(privateViewKey, true);
	//std::cout << "+++myViewKey   : " << myViewKey.toString() << std::endl;
*/


	//const std::string pubAddr = "abc"; //args[2];
	//std::cout << "USING PUBADDR: " << pubAddr << std::endl;
				
	//PubXmrAddress pubXmrAddress(pubAddr);
		
	//const BigInt<257> pubSpend = pubXmrAddress.getPubSpend();
	//const BigInt<257> pubView = pubXmrAddress.getPubView();
		
	//std::cout << "----------------> pubSpend: " << pubSpend << std::endl;
	//std::cout << "----------------> pubView : " << pubView << std::endl;
	
	//const std::string pubSpendStr = pubSpend.getBinary();
	//const std::string pubViewStr = pubView.getBinary();
	
				
	//const std::string convertedAddr = pubXmrAddress.toString();		
		
	//if(pubAddr != convertedAddr)
	//{
	//	throw std::string("Error, addresses does not match");
	//}

	//const std::string xmrPubAddress = convertedAddr;


