/**
 * pubKey.cpp - Bitcoin poscript generator
 *
 * Copyright (C)2015 The bitgen developers <bitgen@openmailbox.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */


#include "pubKey.h"
#include "base58.h"


bool operator<(const PubKey& a, const PubKey& b)
{
	return (a.point.x < b.point.x);
}

bool operator==(const PubKey& a, const PubKey& b)
{
	return (a.point.x == b.point.x);
}



/*
PubKey Pkif::getPubKey() const
{
	bool compressed(false);
	bool onlyIndicatedUsage(false);
	bool signVerify(false);
	
	const ECPoint point = PubKey::parsePkif(pkif, compressed, onlyIndicatedUsage, signVerify);
	PubKey pubkey(point, compressed);
	return pubkey;
}
*/



//Todo: static member
void verifyPkifChecksum(const std::string& pkif, const BigInt<512>& decoded)
{
	const BigInt<512> withoutChecksum = decoded >> 32;

	//Take the last four bytes	
	const BigInt<512> lastFour = decoded.returnMaskHighBits(32);
	
	const BigInt<264> myHash(withoutChecksum); 
	const BigInt<264> xorVal(getPkifAddBeforeHash());	
	const BigInt<264> toHash  = myHash ^ xorVal;
	
	const std::string toHashStr = toHash.getBinaryLong();
	 
	const std::string hashed = doubleSha256(toHashStr);

	const std::string firstFourBytes = hashed.substr(0,4);
	const BigInt<512> checkSum(firstFourBytes, 'c');

	if(lastFour != checkSum)
	{
		std::string errMsg("Incorrect pkif checksum: ");		
		errMsg += pkif;
		errMsg += " Stated value: ";
		errMsg += hexString(lastFour.getBinaryLong());
		errMsg += " Computed value: ";
		errMsg += hexString(checkSum.getBinaryLong());
		
		throw errMsg;
	}	
}







//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!1
const ECPoint //Pkif
PubKey::parsePkif(const std::string& pkif, bool& compressed, 
	bool& onlyIndicatedUsage, bool& signVerify)
{		
	const BigInt<512> fullDecoded = Base58Decode<512>(pkif);	
	verifyPkifChecksum(pkif, fullDecoded);
	
	//Take the last four bytes
	const BigInt<512> withoutChecksum = fullDecoded >> 32;

	const BigInt<512> highByte = withoutChecksum >> 256;

	const BigInt<512> highFourBits = highByte >> 4;
	
	const BigInt<512> highfourBitsShould(getPkifFourBitVal());
	
	if(highFourBits != highfourBitsShould)
	{
		std::cout << "High bits: " << highFourBits << std::endl;
		std::cout << "Should be: " << highfourBitsShould << std::endl;
		
		throw std::string("Incorrect high byte");
	}
	
	onlyIndicatedUsage = highByte[3];
	signVerify = highByte[2];
	compressed = highByte[1];
	const bool parityOdd = highByte[0];
	
	const BigInt<1024> x(withoutChecksum.returnMaskHighBits(256)); 

	BitcoinEllipticCurve ec;
	const ECPoint p = ec.recoverPointFromXAndParityOdd(x, parityOdd);
		
	return p;
}

bool PubKey::isPkif(const std::string& pkif)
{
	try
	{
		bool compressed(false);
		bool onlyIndicatedUsage(false);
		bool signVerify(false);
		
		PubKey::parsePkif(pkif, compressed, onlyIndicatedUsage, signVerify);		
	}
	catch(const std::string& err)
	{
		return false;
	}
	
	return true;
}




ECPoint PubKey::parseVerbosePubKey(const BigInt<1024>& xy)
{
	const BigInt<1024> x = xy >> 256;
	const BigInt<1024> y = xy.returnMaskHighBits(256);
	
	const ECPoint ret(x, y);
	
	BitcoinEllipticCurve ec;
		
	if(!ec.verifyECPoint(ret))
	{
		std::string errStr("Error, point not on curve: ");
		errStr += ret.toString();
		
		throw errStr;
	}
	
	return ret;
	
}









BigInt<264> PubKey::rawPKIFValWithoutChecksum() const
{	
	const bool isOdd = point.y.odd();
	
	const BigInt<256> xVal(point.x);
	
	std::string xStr = xVal.getBinaryLong();
		
	if(xStr.size() != 32)
	{
		throw std::string("Incorrect public key len");
	}
		
	BigInt<264> val(xStr, 'c');	
	
	BigInt<264> num(getPkifFourBitVal());	//Four bit number
	num <<= 260;
	val += num;		//Todo: OR!
	

	//Indicates if encryption of sign verification must be followed
	//0 => Any use permitted
	//1 => Only specifiedy use permitted	
	
	if(onlyIndicatedUsage)
	{
		BigInt<264> compNum(1);
		compNum <<= 259;
		val += compNum;	//Todo: OR!
	}
	
	//Indicates encryption or signature verification	
	//0 => Encrypt
	//1 => Signature verification	
	
	if(signVerify)
	{
		BigInt<264> compNum(1);
		compNum <<= 258;
		val += compNum;	//Todo: OR!
	}	
	
	if(compressed)	
	{
		BigInt<264> compNum(1);
		compNum <<= 257;
		val += compNum;	//Todo: OR!
	}	

	if(isOdd)	
	{
		BigInt<264> compNum(1);
		compNum <<= 256;
		val += compNum;	//Todo: OR!
	}	

	return val;	
}



std::string PubKey::getRawPKIFWithoutChecksum() const
{	
	const BigInt<264> val = rawPKIFValWithoutChecksum();
		
	const std::string stringToMakeHashOf = val.getBinaryLong();
	return stringToMakeHashOf;
}


std::string PubKey::getChecksum() const
{	
	const BigInt<264> val = rawPKIFValWithoutChecksum();

	return computeChecksum(val);
}



std::string PubKey::computeChecksum(const BigInt<264>& val)
{
	const BigInt<264> xorVal(getPkifAddBeforeHash());
	
	const BigInt<264> toHash  = val ^ xorVal;
	
	const std::string stringToMakeHashOf = toHash.getBinaryLong();
	
	const std::string theSha = doubleSha256(stringToMakeHashOf);
		
	const std::string checkSum = theSha.substr(0, 4);

	return checkSum;
}


//Calculates the PKIF
std::string PubKey::createPublicKeyImportFormat() const
{	
	const BigInt<264> val = rawPKIFValWithoutChecksum();	
	const std::string checkSum = computeChecksum(val);
	
	const std::string strVal = val.getBinaryLong();
	
	const std::string cStrWithChecksum = strVal + checkSum;

	const BigInt<296> myval(cStrWithChecksum, 'c');	
	
	const std::string b58Str = Base58Encode(myval);

	return b58Str;
}

