/**
 * xmrEC.cpp - Bitcoin generator
 *
 * Copyright (C)2015 The bitgen developers <bitgen@openmailbox.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
#include "bigint.h"
#include <iostream>

#include "xmrEC.h"




XmrECPoint XmrEllipticCurve::recoverPoint(const BigInt<257>& pubVal)
{
	BigInt<257> pub(pubVal);

	const bool xLsb = pub[255];		//Save bit 255	
	pub.clearBit(255);				//... and clear bit 255

	const BigInt<2048> yval(pub);
	
	const BigInt<2048> xval = recoverX(yval, xLsb);


	const XmrECPoint recoveredPoint(xval, yval);
	
	return recoveredPoint;
}



BigInt<256> XmrEllipticCurve::getEdwards25519Max(const BigInt<256>& privKey)
{
	const BigInt<264> privKeyExtended(BigInt<264>::asPositive(privKey));		

	//Take modulus "l" (the max value for the edwards25519 curve)
	const BigInt<264> privKeyMax(Nval()); //"1000000000000000000000000000000014DEF9DEA2F79CD65812631A5CF5D3ED");
	const BigInt<264> privKeyMod = privKeyExtended % privKeyMax;
 				
	const BigInt<256> privKeyFinal(privKeyMod);
	
	return privKeyFinal;
}









//Calculates (x ^ y) % z
BigInt<2048> XmrEllipticCurve::pow_mod(BigInt<2048> x, const BigInt<2048>& y) const
{
	BigInt<2048> result(1);
			
	const int loopMax = y.numberBitsNeeded();
	
    for(int i = 0 ; i < loopMax ; i++)
    {
		if(y[i])
		{
			result = (result * x) % p;
		}
		
		x = x * x % p;
	}
	        
    return result;
}






BigInt<2048> XmrEllipticCurve::modinv(const BigInt<2048>& inA, const BigInt<2048>& inN) const
{	
	const BigInt<2048> a(inA);
	const BigInt<2048> n(inN);

    BigInt<2048> lm(1);
    BigInt<2048> hm(0);
    
    BigInt<2048> low  = a % n;
    BigInt<2048> high = n;

    
    while(low > one)
    {
        const BigInt<2048> ratio = high / low;
        
        const BigInt<2048> nm = hm - lm * ratio;
        
        const BigInt<2048> theNew = high - low * ratio;
        
        const BigInt<2048> lmCopy(lm);
        const BigInt<2048> lowCopy(low);
        
        lm = nm;
        low = theNew;
        hm = lmCopy;
        high = lowCopy;        
	}

	const BigInt<2048> result(lm % n);
	const BigInt<2048> res(result);
	
    return res;
}



BigInt<2048> XmrEllipticCurve::recoverX(const BigInt<2048>& y, const bool xLsbShould) const
{	
	const BigInt<2048> pPlusThree =  p + three;
	const BigInt<2048> pPlusThreeDivEight =  pPlusThree >> 3;

	const BigInt<2048> pMinusOne =  p - one;	//Todo: Store constants in object?
	const BigInt<2048> pMinusOneDivFour =  pMinusOne >> 2;

		
	const BigInt<2048> xx = (y * y - one) *  modinv(d * y * y + one, p);
	BigInt<2048> x = pow_mod(xx , pPlusThreeDivEight);
  
  
	const BigInt<2048> I = pow_mod(2, pMinusOneDivFour); //(p-1)/4)
  
	if((x * x - xx) % p != 0)
	{
		x = (x * I) % p;
	}
      
	const bool lastBit = x[0];
	if(lastBit != xLsbShould)
	{
		x = p - x;
	}
  
	return x;
}



BigInt<2048> XmrEllipticCurve::toRange(const BigInt<2048>& val) const
{
	if(val.isNegative())
	{
		const BigInt<2048> valPos = -1 * val;
		const BigInt<2048> valMod = valPos % p;
		return (p - valMod) % p;
	}
	else
	{
		return val % p;
	}
}



BigInt<2048> XmrEllipticCurve::makePositive(const BigInt<2048>& val) const
{	
	if(val.isNegative())
	{
		return p + val;
	}
	else
	{
		return val;			
	}
}


XmrECPoint XmrEllipticCurve::addPoints(const XmrECPoint& p1, const XmrECPoint& p2) const
{	
	//std::cout << "add: " << p1.toString() << " and " << pGen.toString() << std::endl;
	
	if(p1.isZero())
	{
		return p2;
	}
	else if(p2.isZero())
	{
		return p1;
	}	
	else if(p1.x == p2.x && p1.y != p2.y)
	{
		const XmrECPoint res(0, 1);
		return res;	
	}
	else
	{	
		const BigInt<2048> invertX = toRange(one + d * p1.x * p2.x * p1.y * p2.y);
		if(invertX == 0)
		{
			//std::cout << "No X inverse found" << std::endl;
			return XmrECPoint(0, 1);
		}
	
		const BigInt<2048> invertY = toRange(one - d * p1.x * p2.x * p1.y * p2.y);
		if(invertY == 0)
		{
			//std::cout << "No Y inverse found: " << p.toString() << " and " << pGen.toString() << std::endl;			
			return XmrECPoint(0, 1);
		}
				
		const BigInt<2048> xNew = ((p1.x * p2.y + p2.x * p1.y) * modinv(one + d * p1.x * p1.y * p2.x * p2.y, p)) % p;		
		const BigInt<2048> yNew = ((p1.y * p2.y + p1.x * p2.x) * modinv(one - d * p1.x * p1.y * p2.x * p2.y, p)) % p;		

		const XmrECPoint res(makePositive(xNew), makePositive(yNew));

		return res;		
	}
}



XmrECPoint XmrEllipticCurve::multiplyPoint(const BigInt<2048>& val, const XmrECPoint& startPoint) const
{
	if(val == zero)
	{
		throw std::string("Private Key cannot be zero");	//Return Point(0,1)??
	}
	
    if(val >= n)
    {
		std::string errMsg("Private Key cannot be larger than N.\n");
		errMsg += val.strHex();
		errMsg += " larger than \n";
		errMsg += n.strHex();
		
		throw errMsg;
	}

	
	XmrECPoint doublingPoint(startPoint);
	
	XmrECPoint res(0, 1);
	
	BigInt<2048> nVal(val);
	
	for(int i = 0 ; ; i++)
	{	
		//std::cout << "Iterating: " << i << std::endl;		
		const BigInt<2048> thisBit = nVal % two;
		nVal /= 2;

		//std::cout << "Doubling point: " << doublingPoint.toString() << std::endl;
		
		if(thisBit == one)
		{
			//std::cout << "Adding: " << i << std::endl;
			res = addPoints(res, doublingPoint);
		}

		if(nVal == zero)
		{
			break;
		}

		doublingPoint = addPoints(doublingPoint, doublingPoint);
	}
	
	return res;
}



XmrECPoint XmrEllipticCurve::multiplyGenPoint(const BigInt<2048>& val) const
{
    if(val == zero)
    {
		throw std::string("Private Key cannot be zero");	//Return Point(0,1)??
	}
	
    if(val >= n)
    {
		std::string errMsg("Private Key cannot be larger than N.\n");
		errMsg += val.strHex();
		errMsg += " larger than \n";
		errMsg += n.strHex();
		
		throw errMsg;
	}

	
	XmrECPoint doublingPoint(genPoint);
	
	XmrECPoint res(0, 1);
	
	BigInt<2048> nVal(val);
	
	for(int i = 0 ; ; i++)
	{	
		//std::cout << "Iterating: " << i << std::endl;		
		const BigInt<2048> thisBit = nVal % two;
		nVal /= 2;

		//std::cout << "Doubling point: " << doublingPoint.toString() << std::endl;
		
		if(thisBit == one)
		{
			//std::cout << "Adding: " << i << std::endl;
			res = addPoints(res, doublingPoint);
		}

		if(nVal == zero)
		{
			break;
		}

		doublingPoint = addPoints(doublingPoint, doublingPoint);
	}
	
	return res;
}





